// Shadow version of algtest.cpp. Contains overloaded functions (namely, which have an int* as their last argument.)
#include<iostream>
#include<iomanip>
#include<string>
#include<cstring>
#include<cmath>
#include"globvar.h" // some variables and function declarations

using namespace std;

#define die(s) {printf(s); exit(1);} // Die, displaying the message (s)...

/* ================= clist definition ==================

   clist is a list of counters, which may be identified via the following list.

   0 - if, &&, ||
   1 - ==, >=, > etc (comparison operators)
   2 - -, -=, +, +=, %
   3 - ++, --
   4 - *, *=
   5 - double, int, ... (variable creations)
//   6 - new (not used)
   6 - number of times ppsumred prevented unnecessary function call
   7 - for (count # loops)
   8 - = (assignment)
   9 - [] (find in array, [][] counts twice)
   10 - [] (write to array, [][] counts twice)
   11 - recpartsum (function call counter)
   12 - ppsum
   13 - ppsumarc
   14 - ppsumred (inline function)
   */

// =====================================================
//  REDUCTION MATRIX FUNCTION (redmat)

/* This function uses a matrix to help reduce reducible
   partial partition sums, into irreducible once. In the
   latter, a partial partition sum is said to be
   IRREDUCIBLE if and only if it evaluates to a greater
   value than any of its contained partial partition
   sums. */

// Function declaration for use in functions to come
double ppsum(short int *, int, int, int, double **, short int **,
	int **, int **, double **, double **, int *);

double ppsumarc(short int *, int, int, int, double **, short int **,
	int **, int **, double **, double **, int *);

// Partial partition sum using reduction matrix (do it inline!)
inline double ppsumred(short int * strand, int strlength, int i /*start*/,
	int j /*length*/, double ** expbdg, short int ** redmat,
	int ** offset, int ** kliststart, double ** matr, double ** matrbow, int * clist)
{
    clist[0]++; // if
    clist[1]++; // >
    clist[9]+=4; // [][]
    if(matr[i][j] > 0) return matr[i][j]; // For redmat[i][2*j] == -1, matr[i][j] = 1 has already been done
    clist[0]++; // if
    clist[1]++; // <
    clist[9]+=2; // [][]
    if(redmat[i][2*j] < -1) // Irreducible?
    { // Then we'll just have to do the calculation
	clist[10]++; // [][]
	clist[12]++; // ppsum
	matr[i][j] = ppsum(strand, strlength, i, j, expbdg,
		redmat, offset, kliststart, matr, matrbow, clist);
	return matr[i][j];
    }

clist[6] += 1; // the ppsumred does what it's meant to do
    clist[5]+=2; // int's
    clist[9]+=4; // [][]
    clist[8]+=2; // =
    int k = redmat[i][2*j];
    int l = redmat[i][2*j+1];
    clist[0]++; // if
    clist[1]++; // >
    clist[9]++; // [][]
    if(matr[k][l] > 0)
    {
	clist[8]++; // =
	clist[9]+=2; // [][]
	clist[10]+=2; // [][]
	matr[i][j] = matr[k][l]; // Would it be good to also write all intermediate reduction ppsums? Or is this only extra work?
	clist[0]++; // if
	clist[1]++; // >
	clist[9]++; // [][]
	if(matrbow[k][l] > 1)
	{
	    clist[2]++; // +=
	    clist[9]+=2; // [][]
	    clist[10]+=2; // [][]
	    matr[i][j] += matrbow[k][l];
	}
	clist[0]++; // if
	clist[1]++; // ==
	clist[9]++; // [][]
	if(matrbow[k][l] == 1)
	{
	    clist[8]++; // =
	    clist[13]++; // ppsumarc
	    clist[10]+=2; // [][]
	    matrbow[k][l] = ppsumarc(strand, strlength, k, l, expbdg,
	       	redmat, offset, kliststart, matr, matrbow, clist);
	    clist[2]++; // +=
	    clist[9]+=2; // [][]
	    clist[10]+=2; // [][]
	    matr[i][j] += matrbow[k][l];
	}
	return matr[i][j];
    }
    else // Haven't done this irreducible -but necessary- ppsum yet, so do it now
    {
	clist[6] -= 1; // This couldn't have been prevented, after all
	clist[8]++; // =
	clist[10]+=2; // [][]
	clist[12]++; // ppsum
	matr[k][l] = ppsum(strand, strlength, k, l, expbdg,
		redmat, offset, kliststart, matr, matrbow, clist);
	clist[8]++; // =
	clist[9]+=2; // [][]
	clist[10]+=2; // [][]
	matr[i][j] = matr[k][l];
	clist[0]++; // if
	clist[1]++; // >
	if(matrbow[k][l] > 1)
	{
	    clist[2]++; // +=
	    clist[9]+=2; // [][]
	    clist[10]+=2; // [][]
	    matr[i][j] += matrbow[k][l];
	}
	clist[0]++; // if
	clist[1]++; // ==
	if(matrbow[k][l] == 1)
	{
	    clist[8]++; // =
	    clist[10]+=2; // [][]
	    clist[13]++; // ppsumarc
	    matrbow[k][l] = ppsumarc(strand, strlength, k, l, expbdg,
	       	redmat, offset, kliststart, matr, matrbow, clist);
	    clist[2]++; // +=
	    clist[9]+=2; // [][]
	    clist[10]+=2; // [][]
	    matr[i][j] += matrbow[k][l];
	}
	return matr[i][j];
    }
    
}

// =====================================================
//   PARTIAL PARTITION SUM WITH START-END ARCS (ppsumarc)

/*  This function calculates a special case of partial
    parition sums, namely, the part of the partial
    partition sum which consists of all configurations
    that have base i (the first) and base i+j-1 (the
    last) connected.

    Call this function ONLY when matrbow[i][j] != 0.
    
    Input parameters:
    strand - The strand of DNA and/or RNA as numbers 0-7
    strlength - length of the complete strand
    start - where the partial sum starts
    length - length of the partial sum (# of nucleotides)
    expbdg - energy contrib. for a bow, depends on config.
    offset - Startpoints of lists of pairs. Required to
    	     know where to end the list of k's.
    kliststart - for every i, contains pointer to first
    		 relevant k in list of k's.
    matr - matrix containing calculated partition sums.
    matrbow - same for partsums with start-end arcing.

    Output parameters:
    result - Return the value obtained for this partition
    	     sum.
    */
// =====================================================

double ppsumarc(short int * strand, int strlength, int start,
	int length, double ** expbdg, short int ** redmat,
	int ** offset, int ** kliststart, double ** matr, double ** matrbow, int * clist)
{
    clist[0]++; // if
    clist[1]++; // >
    if(length<4+MINBOW) return 0; // No bow, so NO contribution!
    clist[5]++; // double
    double innerbow;
    clist[0]++; // if
    clist[1]++; // >
    if(length > MINBOW + 8) // Nontrivial innerbow possible
    {
	clist[8]++; // =
	clist[14]++; // ppsumred
	innerbow = ppsumred(strand, strlength, start+2, length-4, expbdg,
	    redmat, offset, kliststart, matr, matrbow, clist);
    }
    else
    {
	clist[8]++; // =
	innerbow = 1;
    }
    clist[0]++; // if
    clist[1]++; // ==
    clist[9]+=2; // [][]
    if(matrbow[start+1][length-2] == 1)
    {
	clist[13]++; // ppsumarc
	matrbow[start+1][length-2] = ppsumarc(strand, strlength, start+1,
		length-2, expbdg, redmat, offset, kliststart, matr, matrbow, clist);
    }
    clist[2]++; // +=
    clist[9]+=2; // [][]
    innerbow += matrbow[start+1][length-2];
    
    // Return the outcome.
    clist[9]+=4; // [] and [][]
    clist[4]++; // *
    clist[2]++; // +
    return innerbow*expbdg[strand[start]][strand[start+length-2]];
}


// =====================================================
//   PARTIAL PARTITION SUM FUNCTION (ppsum)

/*  This function calculates a partial partition sum.
    It calls itself, and the partial partition sum with
    a start-end arc named ppsumarc.

    Note to self: In case that turns out to decrease
    computation time, a lot of if(matr[i][j] == 1) {...}
    statements may be inserted.
 
    N.B. This result does NOT include start-end arcing.
    For this, use the ppsumarc function.

    Input parameters:
    strand - The strand of DNA and/or RNA as numbers 0-7
    strlength - length of the complete strand
    start - where the partial sum starts
    length - length of the partial sum (# of nucleotides)
    expbdg - energy contrib. for a bow, depends on config.
    offset - Startpoints of lists of pairs. Required to
    	     know where to end the list of k's.
    kliststart - for every i, contains pointer to first
    		 relevant k in list of k's.
    matr - matrix containing calculated partition sums.
    matrbow - same for partsums with start-end arcing.

    Output parameters:
    result - Return the value obtained for this partition
    	     sum.
    */

// =====================================================

double ppsum(short int * strand, int strlength, int start,
	int length, double ** expbdg, short int ** redmat,
	int ** offset, int ** kliststart, double ** matr, double ** matrbow, int * clist)
{
    clist[0]++; // if
    clist[1]++; // >
    clist[9]+=2; // [][]
    if(matr[start][length] > 0.000001) return matr[start][length]; // M(i,j)>0? then return M(i,j)
    clist[5]+=4; // int, int*
    clist[8]+=2; // =
    int k;
    int * ptr;
    int i = start; // Start position of partial partition sum
    int j = length; // Length of partial partition sum (# of nucleotides)
    clist[0]++; // if
    clist[1]++; // <
    if(j<5+MINBOW) {matr[i][j] = 1; clist[8]++; clist[10]+=2; return 1;} // Too short, no arcing possible, or only a start-end arc
    clist[5]+=3;
    double result; // Result to be returned
    double extarc; // Replaces 'prebow'
    double intarc; // Replaces 'innerbow'
    clist[8]++; // =
    clist[14]++; // ppsumred
    result = ppsumred(strand, strlength, i+1, j-1, expbdg,
		redmat, offset, kliststart, matr, matrbow, clist);
    clist[0]++; // if
    clist[1]++; // >
    clist[9]++; // []
    if(strand[i+1] > 7) return result; // If part of strand starts with an X, this is all there is.
    // Find out whether bow has been calc'ed before:
    clist[0]++; // if
    clist[1]++; // ==
    clist[9]++; // [][]
    if(matrbow[i+1][j-1] == 1)
    {
	clist[8]++; // =
	clist[10]+=2; // [][]
	clist[13]++; // ppsumarc
	matrbow[i+1][j-1] = ppsumarc(strand, strlength, i+1, j-1,
		expbdg, redmat, offset, kliststart, matr, matrbow, clist);
    }
    clist[2]++; // +=
    clist[9]+=2; // [][]
    result += matrbow[i+1][j-1];
    clist[0]++; // if
    clist[1]++; // >
    clist[9]++; // []
    if(strand[i] > 7) return result; // There's nothing more because we have one X
    // Initialization done, now do the actual work
    clist[8]++;
    clist[9]++;
  for(ptr=kliststart[i]; ptr<offset[4*(3-strand[i+1]%4)+(3-strand[i]%4)+1]; ptr++)
    {
	clist[7]++; // for
	clist[2]++; // -
	clist[8]++; // =
	clist[9]++; // []
	k = ptr[0]+2-i; // +1 because the k is the first of the pair (in the list of k's).
			// We want an i -> k and i+1 -> k-1 arc.
			// But also +1 because k is the number
			// of positions, k=i_end-i_start+1.
	clist[0]+=3; // three ifs
	clist[1]+=3; // > < ==
	clist[9]+=2; // [][]
	if(k > j-1) break; // Too far, bow exceeds partsum. Recall that partsum ends at position j-1.
	if(k < MINBOW+4) continue; // Too small bow, perhaps next bow is better
	if(matrbow[i][k] == 1)
	{
	    clist[8]++; // =
	    clist[10]+=2; // [][]
	    clist[13]++; // ppsumarc
	    matrbow[i][k] = ppsumarc(strand, strlength, i, k, expbdg,
		redmat, offset, kliststart, matr, matrbow, clist);
	}
	clist[8]++; // =
	clist[9]+=2; // [][]
	intarc = matrbow[i][k];
	clist[0]++; // if
	clist[1]++; // >
	clist[2]++; // -
	if(j - k > 4)//3+MINBOW) // Note that the arc ends at kth pos. s.t.
	    	      // we have positions k+1, ..., j-1 left.
	{
	    clist[2]+=2; // + -
	    clist[8]++; // =
	    clist[14]++; // ppsumred
	    extarc = ppsumred(strand, strlength, i+k, j-k, expbdg,
		redmat, offset, kliststart, matr, matrbow, clist);
	    clist[0]++; // if
	    clist[1]++; // ==
	    clist[9]+=2; // [][]
	    if(matrbow[i+k][j-k] == 1)
	    {
		clist[2]+=2; // + -
		clist[8]++; // =
		clist[10]+=2; // [][]
		clist[13]++; // ppsumarc
		matrbow[i+k][j-k] = ppsumarc(strand, strlength, i+k, j-k, expbdg,
		    redmat, offset, kliststart, matr, matrbow, clist);
	    }
	    clist[2]+=3; // + - +=
	    clist[9]+=2; // [][]
	    extarc += matrbow[i+k][j-k]; // Possibly (often) zero
	}
	else
	{
	    clist[8]++;
	    extarc = 1;
	}
	clist[2]++; // +=
	clist[4]++; // *
	result += intarc*extarc; // Contribution to ppsum
    }
    clist[8]++; // =
    clist[10]++; // [][]
    matr[i][j] = result;
    return result;
	    
} // End ppsum function



// =====================================================
//   PARTITION FUNCTION, v2

/*  This version of the Partition Function is an
    improved version of the previous one. Recursive
    techniques are used for a more efficient calculation.

    Input parameters:
    strand - Strand in DNA and/or RNA translated to ints
    strlength - Length of strand
    bindEnerg - Nearest-Neighbor stacking free energy params
    */
// =====================================================

double recpartsum(short int * strand, int strlength, double ** expbdg, double** matr,
	double** matrbow, int** offset, int** kliststart, short int ** redmat, int cleanmatrcount, int * clist)
{
    clist[5]+=5; // vars
    clist[8]++; // =
    int i,j,k;
    double result = 0; // Store the result for the partition sum here.
    int * ptr; // Pointer as loop variable (Loop through array)
clist[8]++; // =
for(k = 0; k<cleanmatrcount; k++)
{
    clist[3]++; // ++
    clist[1]++; // <
    clist[7]++; // for
    // INITIALIZE PARTIAL RESULT ARRAYS
    clist[8]++; // =
    for(i = 0; i<strlength; i++)
    {
	clist[3]++; // ++
	clist[1]++; // <
	clist[7]++; // for
	
	clist[8]++; // =
	for(j = 0; j<strlength+1; j++)
	{
	    clist[1]++; // <
	    clist[3]++; // ++
	    clist[7]++; // for
	    
	    clist[8]+=2; // =
	    clist[9]+=2; // [][]
	    clist[10]+=2; // [][]
	    matr[i][j] = matrbow[i][j] = 0;
	    clist[0]++; // if
	    clist[1]++; // ==
	    clist[8]++; // =
	    clist[9]+=2; // [i][2*j]
	    clist[10]+=2; // [i][j]
	    if (redmat[i][2*j] == -1) matr[i][j] = 1;
	}
    }
    // For all i, list bows i->k but be careful to use only the right startlist.
    // This list might be empty, s.t. kliststart[i] points to 'last'=offset[16].
    for(i = 0; i<strlength-3; i++)
    {
	clist[7]++;

	clist[0]+=2; // if ||
	clist[9]+=2; // []
	if(strand[i]>7 || strand[i+1]>7) continue; // X here, no offset needed

	clist[8]++;
	clist[9]++;
	for(ptr = kliststart[i]; ptr < offset[4*(3-strand[i+1]%4)+(3-strand[i]%4)+1]; ptr++)
	{
	    clist[1]++; // <
	    clist[2]+=5; // + - %
	    clist[3]++; // ++
	    clist[7]++; // for
	    clist[9]+=3; // []

	    clist[2]++; // -
	    clist[8]++; // =
	    clist[9]++; // []
	    clist[10]+=2; // [][]
	    matrbow[i][ptr[0]-i+2] = 1;
	}
    }
} // end cleanmatrcount loop
    clist[8]++; // =
    clist[14]++; // ppsumred
    result = ppsumred(strand, strlength, 0, strlength, expbdg,
	redmat, offset, kliststart, matr, matrbow, clist); // No-bow partsum
    clist[0]++; // if
    clist[1]++; // ==
    clist[9]+=2; // [][]
    if(matrbow[0][strlength] == 1)
    {
	clist[10]+=2; // [][]
	clist[13]++; // ppsumarc
	matrbow[0][strlength] = ppsumarc(strand, strlength, 0, strlength,
		expbdg, redmat, offset, kliststart, matr, matrbow, clist);
    }
    result += matrbow[0][strlength]; // start-end arc partsum

// ============= FINAL: DELETE DYN.ALL.MEMORY AND QUIT ==========
    
    return result;
    
} // end recursive partition sum

// =====================================================
//   PARTITION FUNCTION (from Partsum_NN.cpp)

/*  The Partition Function will use the known values
    for the pair binding energies to determine the
    total partition sum for a given configuration.
    To determine the Partition Sum, a double sum has to
    be evaluated.
    Function variables:
    * strand: A strand of numbers 0-7 to represent RNA
    * strlength: The length of the Dyn. All. strand
*/
// =====================================================

double partsum(short int * strand, int strlength, double ** expbdg, double ** Partial_PS, double ** Partial_PSbow, int* clist)
{
    clist[0]++; // if
    clist[1]++; // >=
    clist[2]++; // -
    if(MINBOW >= strlength-2) return 1; // There can be no possible connection. But
    		// ... if we continue, terrible things happen, so return 1 right now.
//cout << "partsum starting!\n";
    clist[5] += 6; // allocate vars
    int i, j, k;  // defining the necessary variables for the triple loop
    double betadeltag; // To keep notation short in calculations
    double prebow; // The partition sum ahead of the bow, which may have a bow i->k-1
    double innerbow; // The partition sum inside AND including the bow i+k->i+j
  
/*  Dynamic allocation of the array of doubles that will represent
    partial partition sums. Because of the properties of dynamic
    allocation, it is NOT possible to generate a multidimensional
    array. There are two possible approaches:
    1. Allocate a single-dimensional array with the same number of elements.
    2. Allocate in two steps.
    Notation:
    Partial_PS[i][j] = partial partition sum starting at position [i] of length [j]
    (s.t. it has j+1 nucleotides, and it contains nucleotides i up to i+j.)

    Additionally, dynamic allocation of an array that contains partition sums
    that have their ends connected by a bow. This kind of partition sum is
    required to allow complementary strands longer than 2 base pairs to be
    taken into account correctly. Notation is same as above.
 
 */

/*	As a matter of fact, it's just combinatorics from this point onwards, but
	consequently, this is the hardest part. */
    clist[2]++; // +
    clist[8]++; // =
    for (j=MINBOW+3; j<strlength; j++) // j is the bowlength
    {
	clist[7]++; // for
	clist[1]++; // <
	clist[3]++; // ++
	for (i=0; i < (strlength - j); i++) // i is the starting position of the i-j partition sum under consideration
	{
	    clist[7]++; // for
	    clist[1]++; // <
	    clist[3]++; // ++
	    clist[2]++; // -
	    Partial_PS[i][j] = Partial_PS[i][j-1];	// Start with the No-Bow case, there is NO i->j bow.
	    clist[8]++; // =
	    clist[9]+=2; // [i][j-1]
	    clist[10]+=2; // [i][j]
	    Partial_PS[i][j] += Partial_PSbow[i][j-1]; // Now, we DO want to allow this bow in the partition sum
	    clist[2]++; // +=
	    clist[9]+=2; // [i][j-1]
	    clist[10]+=2; // [i][j]

	    clist[8]++;
	    for (k=0; k<j-2-MINBOW; k++)		// Consider a bow i+k->i+j and i+k+1->i+j-1
	    {
		clist[7]++; // for
		clist[2]++; // -
		clist[3]++; // ++
		clist[1]++; // <
// To fit this on a page, the indent was shifted to the left!

// First, check if the bow is favorable, if nucleotides are not complementary, quit current loop.
// To do this, consider only those combinations whose indices sum up to 3 or 7 or 11.
// Since dAdT = 0+3, dCdG = 1+2, dArU = 0+7, dCrG = 1+6, rArU = 4+7, et cetera,
// and no illegal combination of two elements of the
// set {0,1,2,3,4,5,6,7} gives 3 modulo 4. (Solutions are 0+3,1+2,0+7,1+6,2+5,3+4,4+7,5+6
// which correspond to dAdT,dCdG,dArU,dCrG,dGrC,dTrA,rArU,rCrG. Thus, we have a one-to-one
// mapping between pairs of numbers <8 and NN pairs, designed such that every complementary
// pair sums up to 3 mod 4, while every noncomplementary pair sums up to something else.
		
// Don't do anything unless there is a bow
clist[0]+=2; // if,||
clist[1]+=2; // !=
clist[2]+=8; // +(6), %(2)
clist[9]+=4; // []
if( ((strand[i+k] + strand[i+j]) % 4) != 3 || ((strand[i+k+1] + strand[i+j-1]) % 4) != 3)
    continue; // some things don't match
clist[0]+=2; // if, ||
clist[1]+=2; // >
clist[2]+=6; // +
clist[9]+=4; // []
if( (strand[i+k] + strand[i+j]) > 14 || strand[i+k+1] + strand[i+j-1] > 14)
    continue; // we're having X's here, forbidden

// Calculate the value of the exponent for this bow
clist[2]+=2; // +
clist[8]++; // =
clist[9]+=2; // [][]
betadeltag = expbdg[strand[i+k]][strand[i+j-1]];

// Calculate inner bow and prebow
// Prebow consists of the partition sum (bow or no-bow i->i+k-1) from the start (i) to the bow (i+k-1)
clist[0]++; // if
clist[8]++; // =
if(k>3)
{
    clist[2]++; // +
    clist[9]+=4; // [][]
    prebow = Partial_PS[i][k-1] + Partial_PSbow[i][k-1]; // The TOTAL partition sum for i up to i+k-1
}
else prebow = 1;
// Inner bow consists of (Bow energy) x (Partition sum of i+k+2->i+j-2 with NO bow between the ends)
// 		    +	 (Bow energy) x (Partition sum of i+k+1->i+j-1 with FORCED bow between the ends)
clist[0]++; // if
clist[8]++; // =
if(k<j-3)
{
    clist[2]+=2; // + -
    clist[4]++; // *
    clist[9]+=2; // [][]
    innerbow = betadeltag*Partial_PS[i+k+2][j-k-4];
}
else innerbow = 1;
clist[2]+=3; // += + -
clist[4]++; // *
clist[9]+=2; // [][]
innerbow += betadeltag*Partial_PSbow[i+k+1][j-k-2];
// in the MINBOW=0 case, the "if" statement is necessary to prevent
// segmentation faults. Else, this will not occur.

clist[2]+=2; // + -
clist[8]++; // =
clist[10]+=2; // [][]
Partial_PSbow[i+k][j-k] = innerbow;
clist[0]++; // if
clist[1]++; // >
if(k>0)
{
    clist[2]++; // +=
    clist[4]++; // *
    clist[10]+=2; // [][]
    Partial_PS[i][j] += innerbow*prebow;
}
	    }
	}
    }
    double partsum = Partial_PS[0][strlength-1] + Partial_PSbow[0][strlength-1];

  return partsum;
}

// =====================================================
//   end PARTITION FUNCTION
// =====================================================
