#include<iostream>
#include<fstream>
#include<iomanip>
#include<string>
#include<cstring>
#include<math.h>
#include"sequence.h" // initialize sequences
#include"globdef.h" // some variables and function declarations
//#include"partsum.h"

using namespace std;

#define die(s) {printf(s); exit(1);} // Die, displaying the message (s)...
#define MAXLOOPS 10000 // The number of main loops before terminating this program
#define CONVTOLERANCE 0.000000001 // If delta gets smaller than this, we are very very very near the actual minimum...
#define MAXLENGTH 40 // No combined strand should exceed this size.

inline double modifyexp(double val, double dif)
{
    return exp(log(val)+dif);
}

inline double calcqtotal(double **bindEnerg, double * deltaginit, double * extrapar, double symmterm, double temperature, Sequence * Seqlist, bool output)
{
//    string seq, rcseq, combseq;
    double seqhybr, rcseqhybr, combseqhybr;
    double hybr, nohybr, deltag, deltagnew;
    double qtotal=0;
    double partialresults[3] = {0,0,0};
    int i;
    double * Partial_PSarr = new double[MAXLENGTH*(MAXLENGTH+1)];
    double * Partial_PSbowarr = new double[MAXLENGTH*(MAXLENGTH+1)];
    double ** Partial_PS = new double*[MAXLENGTH];
    double ** Partial_PSbow = new double*[MAXLENGTH];
    for(i = 0; i<MAXLENGTH; i++)
    {
	Partial_PS[i] = Partial_PSarr+i*(MAXLENGTH+1);
	Partial_PSbow[i] = Partial_PSbowarr+i*(MAXLENGTH+1);
    }

    for(i=0; i<SEQLISTLENGTH; i++)
    {
	Seqlist[i].calcpartsum(seqhybr, rcseqhybr, combseqhybr, bindEnerg, Partial_PS, Partial_PSbow, extrapar);
	if(seqhybr > 0 && rcseqhybr > 0 && combseqhybr > 0) {} // If this is not the case, likely something went fatally wrong.
	else
	{
	cout << "seqhybr: " << seqhybr << endl;
	cout << "rcseqhybr: " << rcseqhybr << endl;
	cout << "combseqhybr: " << combseqhybr << endl;
	for(int x = 0; x<8; x++)
	    for(int y = 0; y<8; y++)
		cout << "bindEnerg["<<x<<"]["<<y<<"] = " << log(bindEnerg[x][y]) << endl;
	for(int x = 0; x<23; x++) cout << "extrapar["<<x<<"] = " << log(extrapar[x]) << endl;
	exit(1);
	}
	int type = Seqlist[i].gettype();
	nohybr = seqhybr*rcseqhybr;
	hybr = (combseqhybr - nohybr)*exp(-beta*deltaginit[type]); // Note the minus sign! This sign should be opposite to bindEnerg parameters.
	deltag = log(nohybr/hybr) / beta; // This is the calculated DeltaG for the current parameter set.
	deltag += Seqlist[i].getsymmcheck()*symmterm;
	deltagnew = Seqlist[i].getdeltaG() - (temperature - 310.15)*Seqlist[i].getdeltaS()/1000;
	qtotal += (deltagnew + deltag)*(deltagnew + deltag);
	if(output==1 && i == 63) partialresults[0] = qtotal;
	if(output==1 && i==63+108) partialresults[1] = qtotal - partialresults[0];
	if(output==1 && i==SEQLISTLENGTH-1) partialresults[2] = qtotal - partialresults[1]-partialresults[0];
	if(output) cout << "no " << setw(2) << i+1 << ": exp " << setw(3) << deltagnew << " pred " << setprecision(3) << setw(3) << -deltag << " diff " << setprecision(3) << deltagnew + deltag << endl;
    }
    if(output)
    {
	double referr = 0;
	for(i=0; i<SEQLISTLENGTH; i++)
	    referr += (Seqlist[i].getdeltaG() - Seqlist[i].getdeltaGref())*(Seqlist[i].getdeltaG() - Seqlist[i].getdeltaGref());
	cout << "Reference error: " << setprecision(7) << referr << endl;
	cout << "Partial results were: " << partialresults[0] << " for RNA/DNA, " << partialresults[1] << " for DNA/DNA, and " << partialresults[2] << " for RNA/RNA." << endl;
    }
    delete [] Partial_PS;
    delete [] Partial_PSbow;
    delete [] Partial_PSarr;
    delete [] Partial_PSbowarr;
    if(qtotal < 0 || qtotal > 1000) cout << "qtotal is really large at " << qtotal << ", warning! " << endl;
    return qtotal;
}

void fitdrNN(void)
{
//    string seq, rcseq, combseq;
    Sequence * Seqlist = new Sequence[SEQLISTLENGTH];
    initseqs(Seqlist);
// Calculate the Qtotal from the original paper providing the experimental data and fit
    double refQtot = 0;
    double temperature = 310.15; // Take 37 degrees for temperature
    for(int i = 0; i < SEQLISTLENGTH; i++)
	refQtot += (Seqlist[i].getdeltaG() - Seqlist[i].getdeltaGref())*(Seqlist[i].getdeltaG() - Seqlist[i].getdeltaGref());
    cout << "Reference error: Qtotal = " << refQtot << endl << endl;

    int i,j; // counters

// Parameters in need of fitting
    double deltaginit[3] = {1.0,3.1,4.1};
    const int parcount = 2*7+3*(1+1+1); // 7 hairpin params (1,2,3,4,5,6-7,8+ bases in between), 1 internal AU param, 1 bulge param, 1 external AU param. DNA-DNA, DNA-RNA, RNA-RNA.
    double extrapar[parcount];
    double symmterm = 0.40;
    const int ARRAYSIZE = 8;
    double * bindEnergarr = new double[ARRAYSIZE*ARRAYSIZE];
    double ** bindEnerg = new double*[ARRAYSIZE];
    if(bindEnerg == NULL)
   	die("bindEnerg allocation failed. Abort.");
    for(i=0;i<ARRAYSIZE; i++)
    {
      	bindEnerg[i] = bindEnergarr+i*ARRAYSIZE;
        if(bindEnerg[i] == NULL)
	    die("bindEnerg allocation failed. Abort.");
    }	

    // Initialize extra parameters

    // First: the parameters for DNA hairpins. Only the value extrapar[4] is used (corresponding to a 4-or-more nucleotide loop), so only one hairpin parameter is used.
    extrapar[0] = -8.0;
    extrapar[1] = -6.0;
    extrapar[2] = -4.0;
    extrapar[3] = -3.0; // Antao, Tinoco, 1992: "TPs for loop formation in RNA and DNA hairpin tetraloops"
    extrapar[4] = -4.0;
    extrapar[5] = -5.0;
    extrapar[6] = -6.0;

    // Second: the parameters for RNA hairpins. Using only the parameter extrapar[11].
    extrapar[7] = -8.0;
    extrapar[8] = -5.5;
    extrapar[9] = -3.5;
    extrapar[10] = -2.5; // Antao, Tinoco, 1992. Parameters depend strongly on the actual nucleotides inside the loop, but we do not have enough data to fit that many parameters.
    extrapar[11] = -3.5;
    extrapar[12] = -4.5;
    extrapar[13] = -5.5;

    extrapar[14] = -0.04; // Penalty for closing AU pair, DNA-DNA
    extrapar[15] = -0.04; // DNA-RNA or RNA-DNA
    extrapar[16] = -0.04; // RNA-RNA

    // Bulge parameters. Those are not used in the algorithm at this time, so they are not fitted.
    extrapar[17] = -1.3; // DNA-DNA. Using an averaged value from Tanaka, Kameda, Yamamoto et al, 2004: "DNA-DNA TP based on NN model for single-bulge loop seqs"
    extrapar[18] = -1.3; // Mixed DNA-RNA
    extrapar[19] = -1.3; // RNA-RNA

    extrapar[20] = -0.4; // Penalty for closing AU pair, DNA-DNA
    extrapar[21] = -0.4; // DNA-RNA or RNA-DNA
    extrapar[22] = -0.4; // RNA-RNA


// Initialize bindEnerg array for T=37`C.

// Note that the first index corresponds to the the type of the first nucleotide
// of the front doublet of the arc, while the second index corresponds to the
// type of the first nucleotide of the second doublet of the arc. Consequently,
// bindEnerg[1][3] = bindEnerg[dnaC][dnaT] = Delta G for a dCdA/dGdT base pair doublet.

// Similarly, bindEnerg[2][5] = bindEnerg[dnaG][rnaC] = dGdG/rCrC.

// Last update: January 20, 2009.

bindEnerg[0][0] = 0.88;
bindEnerg[0][1] = 1.28;
bindEnerg[0][2] = 1.44;
bindEnerg[0][3] = 1.00;
bindEnerg[0][4] = 0.9;
bindEnerg[0][5] = 0.9;
bindEnerg[0][6] = 1.1;
bindEnerg[0][7] = 0.2;

bindEnerg[1][0] = 1.28;
bindEnerg[1][1] = 2.17;
bindEnerg[1][2] = 1.84;
bindEnerg[1][3] = 1.45;
bindEnerg[1][4] = 1.8;
bindEnerg[1][5] = 1.7;
bindEnerg[1][6] = 2.9;
bindEnerg[1][7] = 1.6;

bindEnerg[2][0] = 1.44;
bindEnerg[2][1] = 1.84;
bindEnerg[2][2] = 2.24;
bindEnerg[2][3] = 1.30;
bindEnerg[2][4] = 2.1;
bindEnerg[2][5] = 2.1;
bindEnerg[2][6] = 2.7;
bindEnerg[2][7] = 1.5;

bindEnerg[3][0] = 1.00;
bindEnerg[3][1] = 1.45;
bindEnerg[3][2] = 1.30;
bindEnerg[3][3] = 0.58;
bindEnerg[3][4] = 1.0;
bindEnerg[3][5] = 0.9;
bindEnerg[3][6] = 1.3;
bindEnerg[3][7] = 0.6;

bindEnerg[4][0] = 0.9;
bindEnerg[4][1] = 1.8;
bindEnerg[4][2] = 2.1;
bindEnerg[4][3] = 1.0;
bindEnerg[4][4] = 1.10;
bindEnerg[4][5] = 2.08;
bindEnerg[4][6] = 2.24;
bindEnerg[4][7] = 0.93;

bindEnerg[5][0] = 0.9;
bindEnerg[5][1] = 1.7;
bindEnerg[5][2] = 2.1;
bindEnerg[5][3] = 0.9;
bindEnerg[5][4] = 2.08;
bindEnerg[5][5] = 2.36;
bindEnerg[5][6] = 3.26;
bindEnerg[5][7] = 2.11;

bindEnerg[6][0] = 1.1;
bindEnerg[6][1] = 2.9;
bindEnerg[6][2] = 2.7;
bindEnerg[6][3] = 1.3;
bindEnerg[6][4] = 2.24;
bindEnerg[6][5] = 3.26;
bindEnerg[6][6] = 3.42;
bindEnerg[6][7] = 2.35;

bindEnerg[7][0] = 0.2;
bindEnerg[7][1] = 1.6;
bindEnerg[7][2] = 1.5;
bindEnerg[7][3] = 0.6;
bindEnerg[7][4] = 0.93;
bindEnerg[7][5] = 2.11;
bindEnerg[7][6] = 2.35;
bindEnerg[7][7] = 1.33;

for(i = 0; i<8; i++)
    for(j = 0; j<8; j++)
	bindEnerg[i][j] = exp(beta*bindEnerg[i][j]);

for(i = 0; i<parcount; i++)
    extrapar[i] = exp(beta*extrapar[i]);
    double qtotal = 0; // Q is a measure for the total square-deviation with the current set of parameters
    double qtotalnew = 0; // Q after we have "improved" things
    double delta = 10; // The time step. This should not be too big (no convergence) and not too small (long calculations). To aid this, the value of delta will be varied during runtime.
    double bindEnergdif[8][8];
    for(i=0; i<8; i++) for(j=0; j<8; j++) bindEnergdif[i][j] = 0; //deltaHdif[i][j] = deltaSdif[i][j] = 0; // initialize to 0
    double extrapardif[parcount];
    for(i=0; i<parcount; i++) extrapardif[i] = 0;
    double initdif[3] = {0,0,0};
    double symmtermdif = 0; // Also, this is not expected to be temperature-dependent, so it is added to the deltaH part.
    const int NUMVAR = 93; // up to 93 parameters can be stored
    double vel[NUMVAR];
    bool successflag = 0; // Whether or not the last loop succeeded in decreasing Qtotal
    int altcheck = 0; // To check for alternating success/failure which indicates a limit to convergence. When precision limit is reached,
    		      // the program tends to alternate success/failure without ever being able to decrease Qtotal, because the
    		      // qtotalnew > qtotal comparison is flawed when those two are too close. Precision problems in the change of the
    		      // parameters of the model then kill convergence, while no better convergence can ever be achieved.
    int successcount = 0;

    double resultarray[NUMVAR][11]; // Store the results for [parameter][temperature]
    for(i=0; i<NUMVAR; i++) for(j=0; j<11;j++) resultarray[i][j] = 0;
    int k = 0; // define k as another counter

// The SUPER-loop, which involves the former fitdr program but now over several temperatures. 

	// For a number of runs around N=11 to also be able to estimate dS and dH, use this loop:
//for(int n = -10; n<11; n+=2)
	// For a single run at 37 degrees centigrade, use this loop:
for(int n = 0; n<1; n++)
{
temperature = 310.15 + n; // Do calculations for this specific temperature

// Reset initial parameters...

qtotal = qtotalnew = 0;
for(i = 0; i<NUMVAR; i++) vel[i] = 0;
delta = 10;
successflag = 0;
altcheck = 0;
successcount = 0;
for(k = 0; k<MAXLOOPS+1; k++)
{
    cout << "Loop " << k << "... ";
    qtotal = calcqtotal(bindEnerg, deltaginit, extrapar, symmterm, temperature, Seqlist, 0);
    for(i=0; i<3; i++)
    {
	deltaginit[i] = modifyexp(deltaginit[i],0.0001);
	initdif[i] = calcqtotal(bindEnerg, deltaginit, extrapar, symmterm, temperature, Seqlist, 0) - qtotal;
	deltaginit[i] = modifyexp(deltaginit[i],-0.0001);
    }
    symmtermdif = calcqtotal(bindEnerg, deltaginit, extrapar, symmterm + 0.000001, temperature, Seqlist, 0) - qtotal;
    for(i=0; i<parcount; i++)
    {
	for(j=0; j<parcount; j++) if(log(extrapar[i]) < -10 || log(extrapar[i]) > 10) cout << "log(extrapar["<<i<<"]) was " <<log(extrapar[i]) << ". This may be wrong." << endl;
	for(j=0; j<8; j++) for(int m=0; m<8; m++) if(log(bindEnerg[j][m]) < -10 || log(bindEnerg[j][m]) > 10) cout << "log(bindEnerg["<<j<<"]["<<m<<"]) was "<<log(bindEnerg[j][m])<<"."<<endl;
	if(log(extrapar[i]) < -20 || log(extrapar[i]) > 3) {cout << "extrapar["<<i<<"] was " << log(extrapar[i]) << " which was deemed unreasonable. Aborting." << endl; k=MAXLOOPS;}
	extrapar[i] = modifyexp(extrapar[i],0.00001);
	extrapardif[i] = calcqtotal(bindEnerg, deltaginit, extrapar, symmterm, temperature, Seqlist, 0) - qtotal;
	extrapar[i] = modifyexp(extrapar[i],-0.00001);
    }
    fflush(stdout);
    for(i=0; i<8; i++) for(j=0; j<=i; j++)
    {
	bindEnerg[i][j] = modifyexp(bindEnerg[i][j],0.00001);
	bindEnerg[j][i] = modifyexp(bindEnerg[j][i],0.00001);
	bindEnergdif[i][j] = bindEnergdif[j][i] = calcqtotal(bindEnerg, deltaginit, extrapar, symmterm, temperature, Seqlist, 0) - qtotal;
	bindEnerg[i][j] = modifyexp(bindEnerg[i][j],-0.00001);
	bindEnerg[j][i] = modifyexp(bindEnerg[j][i],-0.00001);
    }
    for(i=0; i<8; i++) for(j=0; j<8; j++)
    {
	vel[8*i+j] -= bindEnergdif[i][j]; // Note the minus sign, we found the upward gradient above!
    }
    vel[90] -= initdif[0];
    vel[91] -= initdif[1];
    vel[92] -= initdif[2];
    vel[66] -= symmtermdif;
    for(i=0; i<parcount; i++)
    {
	vel[67+i] -= extrapardif[i];
    }
    for(i=0; i<8; i++) for(j=0; j<8; j++)
    {
	bindEnerg[i][j] = modifyexp(bindEnerg[i][j],delta*vel[8*i+j]);
    }
    deltaginit[0] += delta*vel[90];
    deltaginit[1] += delta*vel[91];
    deltaginit[2] += delta*vel[92];
    symmterm += delta*vel[66];
    for(i=0; i<parcount; i++)
    {
	extrapar[i] = modifyexp(extrapar[i],delta*vel[67+i]);
    }
    qtotalnew = calcqtotal(bindEnerg, deltaginit, extrapar, symmterm, temperature, Seqlist, 0);
    cout <<"Qtotal is now: " << setprecision(20) << qtotalnew << setprecision(6) << ". ";

    // If the new fit is worse, revert to old fit and try a smaller delta
    if(qtotalnew > qtotal)// || stop == 1)
    {
	for(i=0; i<8; i++) for(j=0; j<8; j++)
	    bindEnerg[i][j] = modifyexp(bindEnerg[i][j],-delta*vel[8*i+j]);
	deltaginit[0] -= delta*vel[90];
	deltaginit[1] -= delta*vel[91];
	deltaginit[2] -= delta*vel[92];
	symmterm -= delta*vel[66];
	for(i=0; i<parcount; i++)
	{
	    extrapar[i] = modifyexp(extrapar[i],-delta*vel[67+i]);
	}
	cout << "Qtotal did not decrease, velocity is put to zero!\n";
	for(i=0; i<NUMVAR; i++) vel[i] = 0;
	if(successflag == 0){ delta = delta/2; cout << "Delta was HALVED!!!"; }
	if(successflag == 1) altcheck += 1;
	else altcheck = 0;
	successflag = 0;
	successcount = 0;
    }
    else
    {
	cout << "Success, Qtotal decreased!\n";
	if(successflag == 0) altcheck += 1;
	else altcheck = 0;
	successflag = 1;
	successcount += 1;
    }
    if(successcount == 10){ delta = delta*2; cout << "Delta was DOUBLED!!!";  successcount = 0;}
    if(delta < CONVTOLERANCE){ cout << "Delta is getting very small now!\n Exiting!";}
    if(altcheck > 20) cout << "Alternating behaviour detected, no better convergence is possible. Exiting!\n";

    // These are the requirements for terminating (the first, k=MAXLOOPS, corresponds to a convergence failure, while the latter two correspond to a converged result).
    if(k == MAXLOOPS || delta < CONVTOLERANCE || altcheck > 20)
    {
	cout << "Writing to file \"param.txt\"..." << endl;
    fstream filestr;
    filestr.open("param.txt", ios::out); //fstream::in | fstream::out);
    if(!filestr) die("No output file could be created. Aborting.\n");
    filestr << "Starting output for parameter fit program." << endl;
    filestr << "The total square error was: " << qtotalnew << endl;
    double referr[3] = {0,0,0};
    for(i=0; i<64; i++) referr[0] += (Seqlist[i].getdeltaG() - Seqlist[i].getdeltaGref())*(Seqlist[i].getdeltaG() - Seqlist[i].getdeltaGref());
    for(i=64; i<64+108; i++) referr[1] += (Seqlist[i].getdeltaG() - Seqlist[i].getdeltaGref())*(Seqlist[i].getdeltaG() - Seqlist[i].getdeltaGref());
    for(i=64+108; i<SEQLISTLENGTH; i++) referr[2] += (Seqlist[i].getdeltaG() - Seqlist[i].getdeltaGref())*(Seqlist[i].getdeltaG() - Seqlist[i].getdeltaGref());
    filestr << "The reference square error was: " << referr[0] << " for RNA/DNA, " << referr[1] << " for DNA/DNA and " << referr[2] << "for RNA/RNA." << endl;
    filestr << "deltaginit = " << deltaginit[0] << ", " << deltaginit[1] << ", " << deltaginit[2] << " for DNA-DNA, DNA-RNA and RNA-RNA, respectively.";
    filestr << "symmterm = " << symmterm << endl;
    filestr << "matrix elements:" << setprecision(5) << endl;
    for(i=0; i<8; i++)
    {
	for(j=0; j<8; j++) filestr << " ["<<i<<"]["<<j<<"] = " << setw(10) << log(bindEnerg[i][j])/beta;
	filestr << endl;
    }
    filestr << "Additional parameters:" << endl;
    for(i=0; i<5; i++)
    {
	filestr << "DNA-DNA loop energy for a "<<i+1<<"-nucleotide loop: " << setw(6) << log(extrapar[i]) << " RNA-RNA: " << setw(6) << log(extrapar[7+i]) << endl;
    }
    filestr << "DNA-DNA loop energy for a 6-7-nucleotide loop: " << setw(6) << log(extrapar[5]) << " RNA-RNA: " << setw(6) << log(extrapar[7+5]) << endl;
    filestr << "DNA-DNA loop energy for a "<<8<<"+-nucleotide loop: " << setw(6) << log(extrapar[6]) << " RNA-RNA: " << setw(6) << log(extrapar[7+6]) << endl;

    filestr << "terminal-AU parameters: " << setw(6) << log(extrapar[14]) << setw(6) << log(extrapar[15]) << setw(6) << log(extrapar[16]) << endl;
    filestr << "bulge parameters:       " << setw(6) << log(extrapar[17]) << setw(6) << log(extrapar[18]) << setw(6) << log(extrapar[19]) << endl;
    filestr << "external AU parameters: " << setw(6) << log(extrapar[20]) << setw(6) << log(extrapar[21]) << setw(6) << log(extrapar[22]) << endl;
    filestr.close();

    qtotal = calcqtotal(bindEnerg, deltaginit, extrapar, symmterm, temperature, Seqlist, 1);
    break;
    }
} // end of main loop over k

// Store the results
for(i=0; i<8; i++) for(j=0; j<8; j++) resultarray[8*i+j][ n/2 + 5 ] = log(bindEnerg[i][j])/beta;
resultarray[64][ n/2 + 5 ] = qtotal;
resultarray[65][ n/2 + 5 ] = k;
resultarray[66][ n/2 + 5 ] = symmterm;
resultarray[90][ n/2 + 5 ] = deltaginit[0];
resultarray[91][ n/2 + 5 ] = deltaginit[1];
resultarray[92][ n/2 + 5 ] = deltaginit[2];
for(i=0; i<parcount; i++) resultarray[67+i][ n/2 + 5 ] = log(extrapar[i]);

} // End of super-main loop over temperatures (loop counter n)

// ************************************************************************
// ************************* Display results ******************************
// ************************************************************************

// Results for each temperature will now be displayed.
cout << "Displaying results...\n" << setprecision(4);
for(i=0; i<NUMVAR; i++)
{
    for(j=0; j<11; j++)
	cout << setw(10) << resultarray[i][j];
    cout << endl;
}

cout << "Qtotal was " << qtotal << " after " << k << " loops." << endl;
cout << setprecision(5);
for(i=0; i<8; i++)
{
    for(j=0; j<8; j++)
    {
	cout << " bindE ["<<i<<"]["<<j<<"]= " << setw(7) << log(bindEnerg[i][j])/beta;
    }
    cout << endl;
}

cout << "Finished!\n";

delete [] bindEnerg;
delete [] bindEnergarr;
delete [] Seqlist;
 
} // end fitrrNN function


// Main function.
int main(void)
{
    cout << "Fitting parameters for the extended Nearest-neighbor model...\n";
    fitdrNN();
    cout << "Finished calculations.\n";
}
